#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <fcntl.h>
#include <time.h>
#include <unistd.h>
#include <sys/poll.h>
#include <pthread.h>

#include "arch.h"
#include "totutil.h"

#ifdef DEBUG
long DebugMessages = DB_SLOT1+DB_SLOT3+DB_SLOT4+DB_SLOT5+DB_SLOT2;
#endif

//**************************
//	prompt user
//**************************

char *getuserline (char *prompt, char *ret, int len)
{
	printf ("%s: ", prompt);

	if (fgets (ret, len - 1, stdin)) {
		ret [strlen (ret) - 1] = 0;
		return ret;
	}

	return NULL;
}

//*************************
//	network stuff
//*************************

unsigned int nslookup (char *h)
{
	struct hostent *ho;

	if (!(ho = gethostbyname (h)))
		return 0;

	return *(unsigned int*)*ho->h_addr_list;
}

int connect_tcp (unsigned int IP, short port)
{
	struct sockaddr_in server;
	int Socket;

	if (IP == 0) return -1;

	if ((Socket = socket (AF_INET, SOCK_STREAM, 0)) == -1)
	{
		perror ("Can't create socket");
		exit (1);
	}

	server.sin_family = AF_INET;
	server.sin_port = htons (port);
	server.sin_addr.s_addr = IP;

	if (connect (Socket, (struct sockaddr*)&server,
	             sizeof (struct sockaddr)) == -1)
	{
DPRINTF (DB_SLOT5, "* * * Can't connect: %s\n", strerror (errno));
		return -1;
	}

	return Socket;
}

int connect_tcp (char *host, short port)
{
	return connect_tcp (nslookup (host), port);
}

int listen_tcp (short int *port)
{
	int fd;
	struct sockaddr_in socka;
	struct sockaddr_in porter;
	socklen_t j;

	fd = socket (AF_INET, SOCK_STREAM, 0);

	socka.sin_family = AF_INET;
	socka.sin_port = htons (*port);
	socka.sin_addr.s_addr = htonl (INADDR_ANY);

	if (bind (fd, (struct sockaddr*)&socka, sizeof socka) || listen (fd, 2))
		return -1;

	j = sizeof (porter);
	if (getsockname (fd, (struct sockaddr*)&porter, &j) < 0)
		return -1;

	*port = ntohs (porter.sin_port);

	return fd;
}

int accept_client (int s)
{
	sockaddr_in S;
	socklen_t t;
	pollfd pfd = { s, POLLIN, 0 };
	int i = p_poll (&pfd, 1, -1);

	return (i == 1) ? accept (s, (sockaddr*)&S, &t) : -1;
}

unsigned int peerIP (int fd)
{
	struct sockaddr_in s;
	socklen_t j = sizeof (s);
	getsockname (fd, (struct sockaddr*)&s, &j);
	return s.sin_addr.s_addr;
}

/* kludge */
static short aaport;
static void *answer (void*)
{
	sockaddr_in S;
	socklen_t t;
	int fd = listen_tcp (&aaport), afd;

	if (fd == -1) {
		aaport = -1;
		return NULL;
	}
	afd = accept (fd, (sockaddr*)&S, &t);
	while (aaport);
	close (afd);
	close (fd);
	return NULL;
}
/* kludge end */

bool isitme (char *hostname)
{
	unsigned int IP1 = nslookup (hostname), IP2;
	int fd;

	if (IP1 == 0 || IP1 == inet_addr ("127.0.0.1")) return false;

	if ((fd = connect_tcp (IP1, 7)) == -1) {
		// echo not working? try something else.
		// there must be a better way. if you know patch
		aaport = 0;
		pthread_t t;
		pthread_create (&t, NULL, answer, NULL);
		while (aaport == 0 || aaport == -1) sleep (1);
		fd = connect_tcp (IP1, aaport);
		IP2 = peerIP (fd);
		close (fd);
		aaport = 0;
		pthread_join (t, NULL);
	} else IP2 = peerIP (fd);
	close (fd);

	return IP2 == IP1;
}

//********************************
//	memory allocations
//********************************

static int isdata (void *x)
{
	char automatique;

	return (x DATASTACK (void*)&automatique);
}

int candelete (void *ptr)
{
	static void* dthresh = new char [1];

	if (ptr == NULL) return 0;
	if (dthresh == NULL)
		dthresh = ptr;

	return isdata (ptr) && dthresh SD_DATA ptr;
}

char *StrDup (char *c)
{
	return strcpy (new char [strlen (c) + 1], c);
}

char *iStrDup (char *c)
{
	return (isdata (c)) ? c : StrDup (c);
}

void iStrDelete (char *c)
{
	if (candelete (c)) delete c;
}

//***********************
//	strings
//***********************

// compare the start of <str> with <part>. If match
// return pointer to the rest of <str> else NULL.

char *strpart (char *part, char *str)
{
	int p = strlen (part), s = strlen (str);
	if (s < p) return NULL;
	return (strncmp (str, part, p)) ? NULL : str + p;
}

char *strcasepart (char *part, char *str)
{
	int p = strlen (part), s = strlen (str);
	if (s < p) return NULL;
	return (strncasecmp (str, part, p)) ? NULL : str + p;
}

void Strncpy (char *dest, char *src, size_t s)
{
	strncpy (dest, src, s - 1);
	dest [s - 1] = 0;
}

int tagparse (char *t, char **tags, int lens[], char **ret, int n)
{
	int i;
	char *c;
	for (i = 0; i < n; i++)
		if (!ret [i][0]
		&& (c = strcasepart (tags [i], t))) {
			Strncpy (ret [i], c, lens [i]);
			return i;
		}

	return -1;
}

//*****************************
//	stream extensions
//*****************************

int fputstring (char *s, FILE *f)
{
	fputs (s, f);
	return fputc (0, f);
}

char *fgetstring (int max, FILE *f)
{
	int i = 0;
	char *tmp;

	tmp = (char*) alloca (max);
	while (i < max) {
		tmp [i] = fgetc (f);
		if (tmp [i] == 0) break;
		if (tmp [i] == EOF) return NULL;
		++i;
	}

	if (i == max) {
		do i = fgetc (f);
		while (i != EOF && i != 0);
		tmp [max] = 0;
	}

	return StrDup (tmp);
}

//***********************
//	crypto stuff
//***********************

cryptotype xyc (cryptotype x, cryptotype y)
{
	cryptotype s, t, u;

	s = 1; t = x; u = y;

	while (u) {
		if (u & 1) s = s * t;
		u>>=1;
		t = t*t;
	}

	return s;
}

cryptotype cryptorand ()
{
	return xyc ((rand ()%100+2)|1, rand()) ^ rand ();
}

cryptotype KeyGeneration (cryptotype b, cryptotype X, cryptotype *Key, cryptotype *y)
{
	*y = cryptorand ();

	*Key = xyc (X, *y);

	return xyc (b, *y);
}

//*********************************
//	network date format
//*********************************

static const char month_snames[12][4] = {
	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};
static const char day_snames[7][4] = {
	"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
};

char *network_date (time_t t, char *ret)
{
	struct tm *tms;

	tms = gmtime(&t);

	/* RFC date format; as strftime `%a, %d %b %Y %T GMT` */
	sprintf(ret,
	"%s, %.2d %s %d %.2d:%.2d:%.2d GMT", day_snames[tms->tm_wday],
	tms->tm_mday, month_snames[tms->tm_mon], tms->tm_year + 1900,
	tms->tm_hour, tms->tm_min, tms->tm_sec);

	return ret;
}

//*********************************************
//  zero-copy-networking (userland version)
//	sendfile returns the number of bytes
//*********************************************

zeroCopy::zeroCopy (char *f)
{
	d = NULL;
	success = ZC_NA;
	fd = -1;
	if ((len = filesize (f)) == -1 || (fd = open (f, O_RDONLY)) == -1)
		return;

	success =
	((d = mmap (0, len, PROT_READ, MAP_PRIVATE, fd, 0)) != MAP_FAILED)
		? ZC_OK : ZC_FF;
}

zeroCopy::~zeroCopy ()
{
	if (d && d != MAP_FAILED) munmap (d, len);
	if (fd != -1) close (fd);
}

int filesize (char *fn)
{
	struct stat statbuf;
	if (stat (fn, &statbuf) == -1) return -1;
	return statbuf.st_size;
}

int sendfile (char *f, int fd)
{
	zeroCopy Z (f);

	return (Z.success != ZC_OK) ? 0 : write (fd, Z.d, Z.len);
}

//***************************************
//	zero-copy adler 32
//***************************************

unsigned long adler32_file (char *f)
{
	zeroCopy Z (f);

	return (Z.success == ZC_OK) ? adler32 (1L, Z.d, Z.len) : 0;
}

//***************************************
// 	Base 64
//***************************************
static const char Base64Table [] =
 "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";

char *Encode64 (char *p, char Encoded [])
{
	char *e;

	e = Encoded;

	while (1)
	{
		if (*p == 0) break;
		*e++ = (*p & 252) >> 2;
		*e++ = ((*p & 3) << 4) + ((*++p & 240) >> 4);
		if (*p == 0) 
		{
			*e++ = 64;
			*e++ = 64;
			break;
		}
		*e++ = ((*p & 15) << 2) + ((*++p & 192) >> 6);
		if (*p == 0) {
			*e++ = 64;
			break;
		} else *e++ = (*p & 63);
		p++;
	}
	
	*e = 65;
	
	for (p = Encoded; p <= e ; p++)
		*p = Base64Table [(int) *p];

	return Encoded;
}
